import json
import logging
import os
import tempfile
from typing import Optional, Literal, Any, Union
from urllib.parse import urlparse

import niquests
from pydantic import BaseModel, Field

from ipfabric.tools.shared import raise_for_status
from ipfabric.models.users import User

logger = logging.getLogger("ipfabric")


class Resources(BaseModel):
    cpu: int
    memory: int


class Extension(BaseModel):
    client: Optional[Any] = Field(None, exclude=True)
    id: str
    name: str
    description: Optional[str] = None
    slug: str
    status: str
    type: Optional[Literal["docker-zip", "docker-image"]] = None
    environment_variables: Optional[list[dict[str, str]]] = Field(alias="environmentVariables")
    access_permission: str = Field(alias="accessPermission")
    accessible_by_users: list[str] = Field(alias="accessibleByUsers")
    resources: Resources

    @property
    def users(self) -> list[User]:
        """
        Fetch users with access to the extension.

        Returns:
            List[User]: List of users with access to the extension
        """
        try:
            return [self.client.settings.local_users.users_by_id[user] for user in self.accessible_by_users]
        except:  # noqa: E722, S5754
            return []


class Extensions(BaseModel):
    client: Any = Field(exclude=True)

    @property
    def extensions(self) -> list[Extension]:
        """
        Fetch all extensions.

        Returns:
            List[Extension]: List of all available extensions
        """
        response = raise_for_status(self.client.get("extensions")).json()["extensions"]
        return [Extension(client=self.client, **ext) for ext in response]

    def extension_by_id(self, extension_id: str) -> Extension:
        """
        Fetch a specific extension by ID.

        Args:
            extension_id (str): The ID of the extension to fetch

        Returns:
            Extension: The requested extension
        """
        return Extension(client=self.client, **raise_for_status(self.client.get(f"extensions/{extension_id}")).json())

    def logs_by_extension_id(self, extension_id: str) -> dict:
        """
        Fetch paginated logs for a specific extension.

        Args:
            extension_id (str): The ID of the extension to fetch logs for

        Returns:
            Dict: The paginated logs response
        """
        return raise_for_status(self.client.get(f"extensions/{extension_id}/logs")).json()

    def extension_by_name(self, name: str) -> Extension:
        """
        Fetch a specific extension by name.
        """
        extension = next((ext for ext in self.extensions if ext.name == name), None)
        if extension is None:
            raise ValueError(f"Extension with name '{name}' not found")
        return extension

    def extension_by_slug(self, slug: str) -> Extension:
        """
        Fetch a specific extension by slug.
        """
        extension = next((ext for ext in self.extensions if ext.slug == slug), None)
        if extension is None:
            raise ValueError(f"Extension with slug '{slug}' not found")
        return extension

    @staticmethod
    def _users_permissions(
        access_permission: Literal["anyone", "only-me", "selected-users"] = "anyone",
        accessible_by_users: Optional[list[str]] = None,
    ):
        if access_permission != "selected-users":
            if accessible_by_users:
                logger.warning("Accessible by Users is only allowed with selected-users permission")
            return {"accessPermission": access_permission, "accessibleByUsers": []}
        return {"accessPermission": access_permission, "accessibleByUsers": accessible_by_users}

    def _register_extension(
        self,
        docker_type: Literal["docker-zip", "docker-image"],
        file: bytes,
        name: str,
        slug: str,
        description: str,
        cpu: int = 1,
        memory: int = 512,
        environment_variables: Optional[dict[str, str]] = None,
        access_permission: Literal["anyone", "only-me", "selected-users"] = "anyone",
        accessible_by_users: Optional[list[str]] = None,
    ) -> niquests.Response:
        files = (
            {"file": ("extension.zip", file, "application/zip")}
            if docker_type == "docker-zip"
            else {"file": ("image.tar", file, "application/x-tar")}
        )

        data = {
            "name": name,
            "slug": slug,
            "description": description,
            "resources": json.dumps({"cpu": cpu, "memory": memory}),
            **self._users_permissions(access_permission, accessible_by_users),
        }

        if environment_variables:
            data["environmentVariables"] = [
                {"name": key, "value": value} for key, value in environment_variables.items()
            ]

        return raise_for_status(self.client.post(f"extensions/{docker_type}", files=files, data=data, timeout=300))

    def register_docker_zip(
        self,
        file: bytes,
        name: str,
        slug: str,
        description: str,
        cpu: int = 1,
        memory: int = 512,
        environment_variables: Optional[dict[str, str]] = None,
        access_permission: Literal["anyone", "only-me", "selected-users"] = "anyone",
        accessible_by_users: Optional[list[str]] = None,
    ) -> niquests.Response:
        """
        Register a new extension using zipped source code. Raises an exception if the extension is not registered successfully.

        Args:
            file (bytes): The zipped source code file
            name (str): Name of the extension
            slug (str): Slug for the extension
            description (str): Description of the extension
            cpu (int): Number of CPU cores for the extension
            memory (int): Memory size in MB for the extension
            environment_variables (Optional[Dict[str, str]]): Environment variables for the extension
            access_permission (Literal["anyone", "only-me", "selected-users"]): Access permission for the extension
            accessible_by_users (Optional[List[str]]): List of User IDs with access to the extension
        """
        return self._register_extension(
            "docker-zip",
            file=file,
            name=name,
            slug=slug,
            description=description,
            cpu=cpu,
            memory=memory,
            environment_variables=environment_variables,
            access_permission=access_permission,
            accessible_by_users=accessible_by_users,
        )

    def register_docker_image(
        self,
        file: bytes,
        name: str,
        slug: str,
        description: str,
        cpu: int = 1,
        memory: int = 512,
        environment_variables: Optional[dict[str, str]] = None,
        access_permission: Literal["anyone", "only-me", "selected-users"] = "anyone",
        accessible_by_users: Optional[list[str]] = None,
    ) -> niquests.Response:
        """
        Register a new extension using a docker image. Raises an exception if the extension is not registered successfully.

        Args:
            file (bytes): The tar file of the docker image
            name (str): Name of the extension
            slug (str): Slug for the extension
            description (str): Description of the extension
            cpu (int): Number of CPU cores for the extension
            memory (int): Memory size in MB for the extension
            environment_variables (Optional[Dict[str, str]]): Environment variables for the extension
            access_permission (Literal["anyone", "only-me", "selected-users"]): Access permission for the extension
            accessible_by_users (Optional[List[str]]): List of User IDs with access to the extension
        """
        self._register_extension(
            "docker-zip",
            file=file,
            name=name,
            slug=slug,
            description=description,
            cpu=cpu,
            memory=memory,
            environment_variables=environment_variables,
            access_permission=access_permission,
            accessible_by_users=accessible_by_users,
        )

    def start_extension(self, extension_id: str) -> niquests.Response:
        """
        Start an extension by its ID. Raises an exception if the extension fails to start.

        Args:
            extension_id (str): The ID of the extension to start
        """
        return raise_for_status(self.client.post(f"extensions/{extension_id}/start"))

    def stop_extension(self, extension_id: str) -> niquests.Response:
        """
        Stop an extension by its ID. Raises an exception if the extension fails to stop.

        Args:
            extension_id (str): The ID of the extension to stop
        """
        return raise_for_status(self.client.post(f"extensions/{extension_id}/stop"))

    def unregister_extension(self, extension_id: str) -> niquests.Response:
        """
        Unregister an extension by its ID. Raises an exception if the extension fails to unregister.

        Args:
            extension_id (str): The ID of the extension to unregister
        """
        return raise_for_status(self.client.delete(f"extensions/{extension_id}"))

    def register_from_git_url(
        self,
        git_url: str,
        name: str,
        slug: str,
        description: str,
        branch: str = "main",
        cpu: int = 1,
        memory: int = 512,
        environment_variables: Optional[dict[str, str]] = None,
        access_permission: Literal["anyone", "only-me", "selected-users"] = "anyone",
        accessible_by_users: Optional[list[str]] = None,
    ) -> Union[niquests.Response, None]:
        """
        Register an extension from a Git repository URL. Supports GitHub and GitLab repositories.
        Downloads the repository, creates a ZIP file, and registers it as a docker-zip extension.

        Args:
            git_url (str): URL to the Git repository (GitHub or GitLab)
            name (str): Name of the extension
            slug (str): Slug for the extension
            description (str): Description of the extension
            branch (str): Branch to download (default: "main")
            cpu (int): Number of CPU cores for the extension
            memory (int): Memory size in MB for the extension
            environment_variables (Optional[Dict[str, str]]): Environment variables for the extension
            access_permission (Literal["anyone", "only-me", "selected-users"]): Access permission for the extension
            accessible_by_users (Optional[List[str]]): List of User IDs with access to the extension

        Example:
            client.extensions.register_from_git_url(
                "https://github.com/username/repo",
                "My Extension",
                "my-extension",
                "Description",
                branch="main"
            )
        """
        parsed_url = urlparse(git_url)

        path_parts = parsed_url.path.rstrip(".git").strip("/").split("/")

        if "github.com" in parsed_url.netloc:
            if len(path_parts) != 2:
                raise ValueError("Invalid GitHub URL format")
            user, repo = path_parts
            download_url = f"https://github.com/{user}/{repo}/archive/{branch}.zip"

        elif "gitlab.com" in parsed_url.netloc:
            if len(path_parts) < 2:
                raise ValueError("Invalid GitLab URL format")
            project = path_parts[-1]  # Last part is the project name
            group = "/".join(path_parts[:-1])  # Everything else is the group path
            download_url = f"https://gitlab.com/{group}/{project}/-/archive/{branch}/{project}-{branch}.zip"

        else:
            raise ValueError("Only GitHub and GitLab repositories are supported")

        response = niquests.get(download_url)
        if response.status_code != 200:
            raise niquests.HTTPError(f"Failed to download repository: {response.status_code}")

        with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
            tmp_file.write(response.content)

        try:
            with open(tmp_file.name, "rb") as f:
                resp = self._register_extension(
                    "docker-zip",
                    file=f.read(),
                    name=name,
                    slug=slug,
                    description=description,
                    cpu=cpu,
                    memory=memory,
                    environment_variables=environment_variables,
                    access_permission=access_permission,
                    accessible_by_users=accessible_by_users,
                )
        finally:
            os.unlink(tmp_file.name)
        return resp



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ipfabric.client import IPFClient

import json
from typing import Union, Literal, Annotated, Optional, Any

from pydantic import (
    field_validator,
    BaseModel,
    Field,
    AnyHttpUrl,
    model_serializer,
    AfterValidator,
    SerializationInfo,
    model_validator,
    RootModel,
)

from ipfabric.tools.shared import valid_slug, raise_for_status

AWS_REGIONS = [
    "af-south-1",
    "ap-east-1",
    "ap-northeast-1",
    "ap-northeast-2",
    "ap-northeast-3",
    "ap-south-1",
    "ap-south-2",
    "ap-southeast-1",
    "ap-southeast-2",
    "ap-southeast-3",
    "ap-southeast-4",
    "ap-southeast-5",
    "ap-southeast-7",
    "ca-central-1",
    "ca-west-1",
    "eu-central-1",
    "eu-central-2",
    "eu-north-1",
    "eu-south-1",
    "eu-south-2",
    "eu-west-1",
    "eu-west-2",
    "eu-west-3",
    "il-central-1",
    "me-central-1",
    "me-south-1",
    "mx-central-1",
    "sa-east-1",
    "us-east-1",
    "us-east-2",
    "us-gov-east-1",
    "us-gov-west-1",
    "us-west-1",
    "us-west-2",
]

URL = Annotated[str, AfterValidator(lambda u: str(AnyHttpUrl(u)))]


def serialize_vendor(data: dict[str, Any], api_token: str) -> dict[str, Any]:
    token, username, password = data.get(api_token, None), data.get("username", None), data.get("password", None)
    if token and any([username, password]):
        raise ValueError(f"Cannot combine {api_token} with username and password.")
    [data.pop(_, None) for _ in ["username", "password", api_token] if not data.get(_, None)]
    return data


class BaseVendorAPI(BaseModel):
    """
    base Vendor API config
    """

    slug: str
    comment: str = ""
    isEnabled: bool = True
    vendor_id: Optional[str] = Field(None, alias="id")
    type: str
    respectSystemProxyConfiguration: bool = True

    @field_validator("slug")
    @classmethod
    def check_slug(cls, slug):
        return valid_slug(slug)

    @model_serializer
    def _ser_model(self, info: SerializationInfo) -> dict[str, Any]:
        _ = {
            type(self).model_fields[k].alias or k if info.by_alias else k: v
            for k, v in dict(self).items()
            if v is not None
        }
        return _

    def verify_connection(self, ipf: IPFClient) -> int:
        return raise_for_status(
            ipf.post(
                (
                    "/settings/vendor-api/verify-connection/reverify"
                    if self.vendor_id
                    else "/settings/vendor-api/verify-connection"
                ),
                json=self.model_dump(by_alias=True),
            )
        ).status_code


def check_input(x: BaseVendorAPI, attrs: list[str], token: list[str] = None):
    data, auth = [getattr(x, _) for _ in attrs], ([getattr(x, _) for _ in token] if token else [])
    if not x.vendor_id and token and any(data) and any(auth):
        raise ValueError(f"Cannot combine `{', '.join(attrs)}` with `{', '.join(token)}`.")
    elif not x.vendor_id and token and not (all(data) or all(auth)):
        raise ValueError(f"Required Fields: `{', '.join(attrs)}` or `{', '.join(token)}`.")
    elif not x.vendor_id and not all(data):
        raise ValueError(f"Required Fields: {', '.join(attrs)}")
    return x


class AssumeRole(BaseModel):
    role: str


class AWS(BaseVendorAPI):
    """
    AWS vendor api support
    """

    type: Literal["aws-ec2"] = "aws-ec2"
    apiKey: Optional[str] = None
    apiSecret: Optional[str] = None
    regions: list = Field(default_factory=list)
    assumeRoles: list[Union[str, dict, AssumeRole]] = Field(default_factory=list)

    maxCapacity: int = 50
    maxConcurrentRequests: int = 100000
    refillRate: int = 10
    refillRateIntervalMs: int = 1000

    @model_validator(mode="after")
    def _check_model(self):
        return check_input(self, ["apiKey", "apiSecret", "regions"])

    @field_validator("regions")
    @classmethod
    def check_region(cls, regions):
        for r in regions:
            if r.lower() not in AWS_REGIONS:
                raise ValueError(f"{r} is not a valid AWS Region")
        return [r.lower() for r in regions]

    @field_validator("assumeRoles")
    @classmethod
    def check_roles(cls, roles):
        validated_roles = []
        for role in roles:
            if isinstance(role, str):
                validated_roles.append(AssumeRole(role=role))
            elif isinstance(role, dict):
                if "role" in role:
                    validated_roles.append(AssumeRole(**role))
                else:
                    raise SyntaxError(f'Role {role} not in \'{{"role": "<arn:aws:iam::*****:role/*****>"}}\' format.')
            elif isinstance(role, AssumeRole):
                validated_roles.append(role)
        return validated_roles


class Azure(BaseVendorAPI):
    """
    Azure vendor api support
    """

    type: Literal["azure"] = "azure"
    clientId: Optional[str] = None
    clientSecret: Optional[str] = None
    subscriptionIds: list[str] = Field(default_factory=list)
    tenantId: str

    maxConcurrentRequests: int = 1000
    maxCapacity: int = 300
    refillRate: int = 10
    refillRateIntervalMs: int = 1000

    @model_validator(mode="after")
    def _check_model(self):
        return check_input(self, ["clientId", "clientSecret", "tenantId"])


class CheckPoint(BaseVendorAPI):
    domains: list[str] = Field(default_factory=list)
    type: Literal["checkpoint-mgmt-api"] = "checkpoint-mgmt-api"
    rejectUnauthorized: bool = True
    apiKey: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    baseUrl: URL

    maxConcurrentRequests: int = 1000
    maxCapacity: int = 300
    refillRate: int = 10
    refillRateIntervalMs: int = 1000

    @model_validator(mode="after")
    def _check_model(self):
        return check_input(self, ["username", "password"], ["apiKey"])

    @model_serializer
    def _ser_model(self, info: SerializationInfo) -> dict[str, Any]:
        return serialize_vendor(super()._ser_model(info), "apiKey")


class CiscoAPIC(BaseVendorAPI):
    """
    Cisco APIC vendor api support
    """

    type: Literal["ciscoapic"] = "ciscoapic"
    rejectUnauthorized: bool = True
    username: Optional[str] = None
    password: Optional[str] = None
    baseUrl: URL

    maxConcurrentRequests: int = 10
    maxCapacity: int = 300
    refillRate: int = 10
    refillRateIntervalMs: int = 1000

    @model_validator(mode="after")
    def _check_model(self):
        return check_input(self, ["username", "password"])


class CiscoFMC(BaseVendorAPI):
    type: Literal["ciscofmc"] = "ciscofmc"
    rejectUnauthorized: bool = True
    apiToken: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    baseUrl: URL

    maxConcurrentRequests: int = 9
    maxCapacity: int = 110
    refillRate: int = 110
    refillRateIntervalMs: int = 60000

    @model_validator(mode="after")
    def _check_model(self):
        return check_input(self, ["username", "password"], ["apiToken"])

    @model_serializer
    def _ser_model(self, info: SerializationInfo) -> dict[str, Any]:
        return serialize_vendor(super()._ser_model(info), "apiToken")


class ForcePoint(BaseVendorAPI):
    """
    ForcePoint vendor api support
    """

    type: Literal["forcepoint"] = "forcepoint"
    rejectUnauthorized: bool = True
    authenticationKey: Optional[str] = None
    baseUrl: URL

    maxConcurrentRequests: int = 10
    maxCapacity: int = 300
    refillRate: int = 10
    refillRateIntervalMs: int = 1000

    @model_validator(mode="after")
    def _check_model(self):
        return check_input(self, ["authenticationKey"])


class GCPCredentials(BaseModel):
    """
    GCP Credentials
    """

    auth_provider_x509_cert_url: str
    auth_uri: str
    client_email: str
    client_id: str
    client_x509_cert_url: str
    private_key: str
    private_key_id: str
    project_id: str
    token_uri: str
    type: str
    universe_domain: str


class GCP(BaseVendorAPI):
    """
    GCP vendor api support

    Args:
        credentialsJson: JSON File, Dictionary, or String of GCP Credentials
    """

    credentialsJson: Optional[Union[GCPCredentials, dict, str]] = None
    type: Literal["gcp"] = "gcp"

    @model_validator(mode="after")
    def _check_model(self):
        return check_input(self, ["credentialsJson"])

    @field_validator("credentialsJson")
    @classmethod
    def validate_credentials(cls, credentials):
        if isinstance(credentials, dict):
            return GCPCredentials(**credentials)
        elif isinstance(credentials, str):
            try:
                with open(credentials, "r") as f:
                    return GCPCredentials(**json.load(f))
            except FileNotFoundError:
                try:
                    return GCPCredentials(**json.loads(credentials))
                except json.JSONDecodeError:
                    raise ValueError("credentialsJson must be a valid JSON file, JSON string or Python dictionary.")
        else:
            return credentials


class JuniperMist(BaseVendorAPI):
    """
    Juniper Mist vendor api support
    """

    type: Literal["juniper-mist"] = "juniper-mist"
    apiVer: Literal["v1"] = "v1"
    baseUrl: str = "https://api.mist.com"
    rejectUnauthorized: bool = True
    apiToken: Optional[str] = None
    baseUrl: URL

    maxConcurrentRequests: int = 250
    maxCapacity: int = 300
    refillRate: int = 10
    refillRateIntervalMs: int = 1000

    @model_validator(mode="after")
    def _check_model(self):
        return check_input(self, ["apiToken"])


class Meraki(BaseVendorAPI):
    """
    Meraki v1 vendor api support
    """

    type: Literal["meraki"] = "meraki"
    organizations: list[str] = Field(default_factory=list)
    rejectUnauthorized: bool = True
    apiVer: Literal["v1"] = "v1"
    apiKey: Optional[str] = None
    baseUrl: URL

    maxCapacity: int = 10
    maxConcurrentRequests: int = 100000
    refillRate: int = 10
    refillRateIntervalMs: int = 1000

    @model_validator(mode="after")
    def _check_model(self):
        return check_input(self, ["apiKey"])


class Prisma(BaseVendorAPI):
    """
    Prisma vendor api support
    """

    type: Literal["prisma"] = "prisma"
    rejectUnauthorized: bool = True
    username: Optional[str] = None
    password: Optional[str] = None
    tsgid: Optional[str] = None

    maxConcurrentRequests: int = 4
    maxCapacity: int = 300
    refillRate: int = 10
    refillRateIntervalMs: int = 1000

    @model_validator(mode="after")
    def _check_model(self):
        return check_input(self, ["username", "password", "tsgid"])


class RuckusVirtualSmartZone(BaseVendorAPI):
    """
    Ruckus Virtual SmartZone vendor api support
    """

    apiVersion: Literal["v9_1"] = "v9_1"
    type: Literal["ruckus-vsz"] = "ruckus-vsz"
    rejectUnauthorized: bool = True
    username: Optional[str] = None
    password: Optional[str] = None
    baseUrl: URL

    maxConcurrentRequests: int = 10
    maxCapacity: int = 300
    refillRate: int = 10
    refillRateIntervalMs: int = 1000

    @model_validator(mode="after")
    def _check_model(self):
        return check_input(self, ["username", "password"])


class SilverPeak(BaseVendorAPI):
    """
    SilverPeak vendor api support
    """

    type: Literal["silverpeak"] = "silverpeak"
    rejectUnauthorized: bool = True
    apiKey: Optional[str] = None
    loginType: Optional[Literal["Local", "RADIUS", "TACACS+"]] = None
    username: Optional[str] = None
    password: Optional[str] = None
    baseUrl: URL

    maxConcurrentRequests: int = 5
    maxCapacity: int = 300
    refillRate: int = 10
    refillRateIntervalMs: int = 1000

    @model_validator(mode="after")
    def _check_model(self):
        return check_input(self, ["username", "password", "loginType"], ["apiKey"])

    @model_serializer
    def _ser_model(self, info: SerializationInfo) -> dict[str, Any]:
        _ = serialize_vendor(super()._ser_model(info), "apiKey")
        if self.apiKey or not self.loginType:
            del _["loginType"]
        elif self.vendor_id and not self.loginType and not self.apiKey:
            raise ValueError("loginType must be specified with username and password.")
        return _


class VeloCloud(BaseVendorAPI):
    """
    VeloCloud vendor api support
    """

    type: Literal["velocloud"] = "velocloud"
    rejectUnauthorized: bool = True
    apiToken: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    baseUrl: URL

    maxConcurrentRequests: int = 1000
    maxCapacity: int = 1000
    refillRate: int = 1000
    refillRateIntervalMs: int = 5000

    @model_validator(mode="after")
    def _check_model(self):
        return check_input(self, ["username", "password"], ["apiToken"])

    @model_serializer
    def _ser_model(self, info: SerializationInfo) -> dict[str, Any]:
        return serialize_vendor(super()._ser_model(info), "apiToken")


class Versa(BaseVendorAPI):
    """
    Versa vendor api support
    """

    type: Literal["versa"] = "versa"
    combinedDiscovery: bool = True
    rejectUnauthorized: bool = True
    username: Optional[str] = None
    password: Optional[str] = None
    baseUrl: URL

    maxConcurrentRequests: int = 10
    maxCapacity: int = 100
    refillRate: int = 2
    refillRateIntervalMs: int = 1000

    @model_validator(mode="after")
    def _check_model(self):
        return check_input(self, ["username", "password"])


class Viptela(BaseVendorAPI):
    """
    Viptela vendor api support
    """

    type: Literal["viptela"] = "viptela"
    combinedDiscovery: bool = True
    rejectUnauthorized: bool = True
    username: Optional[str] = None
    password: Optional[str] = None
    baseUrl: URL

    maxConcurrentRequests: int = 25
    maxCapacity: int = 300
    refillRate: int = 10
    refillRateIntervalMs: int = 1000

    @model_validator(mode="after")
    def _check_model(self):
        return check_input(self, ["username", "password"])


class NSXT(BaseVendorAPI):
    """
    NSXT vendor api support
    """

    type: Literal["nsxT"] = "nsxT"
    username: Optional[str] = None
    password: Optional[str] = None
    baseUrl: URL

    maxConcurrentRequests: int = 40
    maxCapacity: int = 100
    refillRate: int = 50
    refillRateIntervalMs: int = 1000
    rejectUnauthorized: bool = True

    @model_validator(mode="after")
    def _check_model(self):
        return check_input(self, ["username", "password"])


VENDOR_API = Annotated[
    Union[
        AWS,
        Azure,
        CheckPoint,
        CiscoAPIC,
        CiscoFMC,
        ForcePoint,
        GCP,
        JuniperMist,
        Meraki,
        Prisma,
        RuckusVirtualSmartZone,
        SilverPeak,
        VeloCloud,
        Versa,
        Viptela,
        NSXT,
    ],
    Field(discriminator="type"),
]

VENDOR_API_MODELS = (
    AWS,
    Azure,
    CheckPoint,
    CiscoAPIC,
    CiscoFMC,
    ForcePoint,
    GCP,
    JuniperMist,
    Meraki,
    Prisma,
    RuckusVirtualSmartZone,
    SilverPeak,
    VeloCloud,
    Versa,
    Viptela,
    NSXT,
)

TYPE_TO_MODEL = {c.model_fields["type"].default: c for c in VENDOR_API_MODELS}


class VendorAPIModel(RootModel):
    root: VENDOR_API

    def __getattr__(self, name: str) -> Any:
        return getattr(self.root, name)



from typing import Any, Optional, Literal, Annotated

from pydantic import BaseModel, Field, field_validator, model_validator, model_serializer, AfterValidator

from ipfabric.tools.shared import validate_ip_network_str


def valid_networks(v: list[str]) -> list[str]:
    return [validate_ip_network_str(_, ipv6=True) for _ in v]


VALID_NETWORKS = Annotated[list[str], AfterValidator(valid_networks)]


class BaseSetting(BaseModel):
    @model_serializer
    def _ser_model(self) -> dict[str, Any]:
        if not (_ := {v.alias or k: getattr(self, k) for k, v in self.model_fields.items()})["id"]:
            del _["id"]
        return _


class BaseEnableSetting(BaseModel):
    enabled: bool

    @model_serializer
    def _ser_model(self) -> dict:
        return dict(self) if self.enabled else {"enabled": self.enabled}


class Community(BaseSetting, BaseModel):
    communityList: list[str]
    isVrfStd: bool
    protocol: Literal["IPv4", "IPv6"]
    sn: str
    vrf: Optional[str] = None
    community_id: Optional[str] = Field(None, alias="id")

    @model_validator(mode="after")
    def _verify_vrf(self):
        if self.vrf and self.isVrfStd:
            raise ValueError("Cannot provide both custom VRF and set `isVrfStd` to True.")
        if not self.vrf and not self.isVrfStd:
            raise ValueError("You must provide a custom VRF or set `isVrfStd` to True.")
        return self


class ManualLink(BaseSetting, BaseModel):
    dstDeviceSn: str
    dstInterface: str
    type: Literal["L1", "L2"]
    srcDeviceSn: str
    srcInterface: str
    link_id: Optional[str] = Field(None, alias="id")


class Networks(BaseModel):
    exclude: VALID_NETWORKS = Field(default_factory=list)
    include: VALID_NETWORKS = Field(default_factory=list)
    whitelist: VALID_NETWORKS = Field(default_factory=list)

    @model_validator(mode="after")
    def _verify_include_not_empty(self):
        if not self.include:
            raise ValueError("Discovery Settings Network Include list cannot be empty.")
        return self


class SeedList(BaseModel):
    seeds: list[str] = Field(default_factory=list)

    @field_validator("seeds")
    @classmethod
    def _verify_valid_networks(cls, v: list[str]) -> list[str]:
        return [validate_ip_network_str(_, max_size=24, max_v6_size=119, ipv6=True) for _ in v]

    @model_serializer
    def _ser_model(self) -> list[str]:
        return self.seeds


class Tacacs(BaseModel):
    retry: int
    delay: int


class CliRetryLimit(BaseModel):
    authFail: int
    default: int
    tacacs: Tacacs


class CliSessionsLimit(BaseEnableSetting, BaseModel):
    limit: int


class DiscoveryTask(BaseSetting, BaseModel):
    disabled_task_id: Optional[str] = Field(None, alias="id")
    vendor: Optional[str] = None
    family: Optional[str] = None
    platform: Optional[str] = None
    model: Optional[str] = None
    version: Optional[str] = None
    sn: Optional[str] = None
    taskId: str


class BGPLimit(BaseModel):
    enabled: bool
    threshold: int


class LimitDiscoveryTasks(BaseModel):
    alreadyDiscovered: bool
    sourceOfTasks: list[Literal["aciEndpoint", "arp", "routes", "trace", "xdp"]]


class ResolveNames(BaseModel):
    discoveryDevices: bool


class Scanner(BaseEnableSetting, BaseModel):
    shortestMask: int

    @field_validator("shortestMask")
    @classmethod
    def _verify_shortest_mask(cls, v: int) -> int:
        if v > 32 or v < 16:
            raise ValueError("Shortest mask must be number from 16 to 32.")
        return v


class Timeouts(BaseModel):
    authentication: int
    command: int
    login: int
    session: int


class Traceroute(BaseModel):
    port: Optional[int] = None
    protocol: Literal["tcp", "udp", "icmp"]
    scope: VALID_NETWORKS

    @model_validator(mode="after")
    def _verify_port(self):
        if self.protocol == "icmp" and self.port:
            self.port = None
        elif self.protocol != "icmp" and not self.port:
            raise ValueError("Port must be provided if protocol is not icmp.")
        return self


class Ports(BaseModel):
    excludeSubnets: VALID_NETWORKS = Field(default_factory=list)
    subnets: VALID_NETWORKS = Field(default_factory=list)
    label: str = ""
    protocol: Literal["ssh", "telnet"]
    port: int

    @model_validator(mode="after")
    def _verify_subnets_not_empty(self):
        if not self.subnets:
            raise ValueError("Custom Port Network Include list cannot be empty.")
        return self


class DiscoveryHistorySeeds(BaseModel):
    afterDate: Optional[str] = Field(None, pattern=r"^\d{4}-\d{2}-\d{2}$")
    daysLimit: Optional[int] = Field(None, ge=0, le=365)
    enabled: bool = False

    @model_validator(mode="after")
    def _verify_limit(self):
        if not self.enabled and self.afterDate and self.daysLimit:
            raise ValueError("Either daysLimit or afterDate must be defined, but not both.")
        return self



from .rules import TableRule, ConfigRule, InventoryRule
from .sync import Config, MultiConfig, load_config, IPFabric
from .values import TableValue, ConfigValue

__all__ = [
    "Config",
    "MultiConfig",
    "load_config",
    "IPFabric",
    "TableRule",
    "ConfigRule",
    "TableValue",
    "ConfigValue",
    "InventoryRule",
]



import logging
from typing import Any, ClassVar, Union, Optional, List

from pydantic import Field, BaseModel, model_serializer

from ipfabric.models.vendor_api import VENDOR_API, VendorAPIModel, TYPE_TO_MODEL
from ipfabric.tools.shared import raise_for_status


CONNECTION_PARAMS = [
    "rejectUnauthorized",
    "respectSystemProxyConfiguration",
    "maxConcurrentRequests",
    "maxCapacity",
    "refillRate",
    "refillRateIntervalMs",
]

logger = logging.getLogger("ipfabric")


class VendorAPI(BaseModel):
    client: Any = Field(exclude=True)
    snapshot_id: Optional[str] = None
    vendor_api: Optional[List[VENDOR_API]] = None
    _api_url: ClassVar[str] = "settings/vendor-api"

    def model_post_init(self, context: Any, /) -> None:
        if not self.vendor_api:
            self.vendor_api = self.get_vendor_apis()

    @model_serializer
    def _ser_model(self) -> list[dict[str, Any]]:
        return [_.model_dump(by_alias=True) for _ in self.vendor_api]

    @property
    def _settings_url(self) -> str:
        return f"settings/{self.snapshot_id}" if self.snapshot_id else "settings"

    def get_vendor_apis(self) -> List[VENDOR_API]:
        """
        Get all vendor apis and sets them in the Authentication.apis
        :return: self.credentials
        """
        return [
            VendorAPIModel(**_).root for _ in raise_for_status(self.client.get(self._settings_url)).json()["vendorApi"]
        ]

    def add_vendor_api(self, api: VENDOR_API) -> VENDOR_API:
        return VendorAPIModel(
            **raise_for_status(self.client.post(self._api_url, json=api.model_dump(by_alias=True))).json()
        ).root

    @staticmethod
    def _return_api_id(api_id: Union[dict, str, int, VENDOR_API]) -> str:
        if isinstance(api_id, dict):
            api_id = api_id["id"]
        elif isinstance(api_id, (str, int)):
            api_id = str(api_id)
        else:
            api_id = api_id.vendor_id
        return api_id

    def delete_vendor_api(self, api_id: Union[dict, str, int, VENDOR_API]) -> int:
        return raise_for_status(self.client.delete(self._api_url + "/" + self._return_api_id(api_id))).status_code

    def enable_vendor_api(self, api_id: Union[dict, str, int, VENDOR_API]) -> int:
        return raise_for_status(
            self.client.patch(self._api_url + "/" + self._return_api_id(api_id), json={"isEnabled": True})
        ).status_code

    def disable_vendor_api(self, api_id: Union[dict, str, int, VENDOR_API]) -> int:
        return raise_for_status(
            self.client.patch(self._api_url + "/" + self._return_api_id(api_id), json={"isEnabled": False})
        ).status_code

    def verify_connection(self, api: VENDOR_API) -> int:
        return raise_for_status(
            self.client.post(
                (
                    "/settings/vendor-api/verify-connection/reverify"
                    if api.vendor_id
                    else "/settings/vendor-api/verify-connection"
                ),
                json=self.model_dump(by_alias=True),
            )
        ).status_code

    def update_vendor_api(  # TODO: NIM-19186 FIX ME
        self,
        current: Union[VENDOR_API, dict],
        update: Union[VENDOR_API, dict],
        restore_conn_defaults: bool = False,
    ) -> VENDOR_API:
        current = (VendorAPIModel(**current) if isinstance(current, dict) else current).model_dump(by_alias=True)

        if not isinstance(update, dict):
            update = update.model_dump(by_alias=True)
        update.pop("id", None)

        new = {**current, **update}
        if restore_conn_defaults:
            default = {
                k: v
                for k, v in vars(TYPE_TO_MODEL[current["type"]].model_construct()).items()
                if k in CONNECTION_PARAMS
            }
            new.update(default)
        params = {
            "vendorApi": [
                _.model_dump(by_alias=True) if _.vendor_id != current["id"] else new for _ in self.get_vendor_apis()
            ]
        }

        return VendorAPIModel(
            **raise_for_status(self.client.patch(self._settings_url, json=params)).json()["vendorApi"]
        ).root



import logging
from typing import Optional, Dict, Any, Union, overload, Literal

import deepdiff
from pydantic import BaseModel, Field, computed_field

from ipfabric.tools.shared import create_filter

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None

logger = logging.getLogger("ipfabric")

IGNORE_COLUMNS = {"id"}
EXPORT_FORMAT = Literal["json", "csv", "df"]


class BaseTable(BaseModel):
    """model for table data"""

    endpoint: str
    client: Any = Field(None, exclude=True)

    @computed_field
    @property
    def summary(self) -> Union[None, str]:
        if self.client:
            return self.client.oas[self.endpoint].post.summary
        return None

    @computed_field
    @property
    def description(self) -> Union[None, str]:
        if self.client:
            return self.client.oas[self.endpoint].post.description
        return None

    @property
    def name(self):
        return self.endpoint.split("/")[-1]

    @property
    def abs_endpoint(self):
        return self.endpoint if self.endpoint[0] == "/" else "/" + self.endpoint

    @overload
    def fetch(
        self,
        export: Literal["json"] = ...,
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        api_version: Optional[Union[str, int]] = None,
    ) -> list[dict]: ...

    @overload
    def fetch(
        self,
        export: Literal["csv"],
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        csv_tz: Optional[str] = None,
        # api_version: Optional[Union[str, int]] = None,  # TODO: NIM-21720
    ) -> bytes: ...

    @overload
    def fetch(
        self,
        export: Literal["df"],
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        api_version: Optional[Union[str, int]] = None,
    ) -> DataFrame: ...

    def fetch(
        self,
        export: EXPORT_FORMAT = "json",
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        csv_tz: Optional[str] = None,
        api_version: Optional[Union[str, int]] = None,
    ):
        """Gets all data from corresponding endpoint

        Args:
            export: str: Export format to return [json, csv]; default is json.
            columns: Optional columns to return, default is all
            filters: Optional filters'
            attr_filters: dict: Optional dictionary of Attribute filters
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            limit: Default to 1,000 rows
            start: Starts at 0
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
            api_version: Optional API version to use for this request's X-API-Version header,
                         default None will use latest version. Values other than None will not use streaming requests
                         and will switch to pagination. API Version is not supported with CSV export.
        Returns:
            Union[list[dict], bytes, pd.DataFrame]: List of dict if json, bytes string if CSV, DataFrame if df
        """
        return self.client.fetch(
            self.endpoint,
            export=export,
            columns=columns,
            filters=filters,
            attr_filters=attr_filters,
            sort=sort,
            limit=limit,
            start=start,
            snapshot=False,
            csv_tz=csv_tz,
            api_version=api_version,
        )

    @overload
    def all(
        self,
        export: Literal["json"] = ...,
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        sort: Optional[dict] = None,
        api_version: Optional[Union[str, int]] = None,
    ) -> list[dict]: ...

    @overload
    def all(
        self,
        export: Literal["csv"],
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        sort: Optional[dict] = None,
        csv_tz: Optional[str] = None,
        # api_version: Optional[Union[str, int]] = None,  # TODO: NIM-21720
    ) -> bytes: ...

    @overload
    def all(
        self,
        export: Literal["df"],
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        sort: Optional[dict] = None,
        api_version: Optional[Union[str, int]] = None,
    ) -> DataFrame: ...

    def all(
        self,
        export: EXPORT_FORMAT = "json",
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        sort: Optional[dict] = None,
        csv_tz: Optional[str] = None,
        api_version: Optional[Union[str, int]] = None,
    ):
        """Gets all data from corresponding endpoint

        Args:
            export: str: Export format to return [json, csv]; default is json.
            columns: Optional columns to return, default is all
            filters: Optional filters
            attr_filters: dict: Optional dictionary of Attribute filters
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
            api_version: Optional API version to use for this request's X-API-Version header,
                         default None will use latest version. Values other than None will not use streaming requests
                         and will switch to pagination. API Version is not supported with CSV export.
        Returns:
            Union[list[dict], bytes, pd.DataFrame]: List of dict if json, bytes string if CSV, DataFrame if df
        """
        return self.client.fetch_all(
            self.endpoint,
            export=export,
            columns=columns,
            filters=filters,
            attr_filters=attr_filters,
            sort=sort,
            snapshot=False,
            csv_tz=csv_tz,
            api_version=api_version,
        )

    def count(
        self,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        api_version: Optional[Union[str, int]] = None,
    ) -> int:
        """
        Gets count of table
        :param filters: dict: Optional filters
        :param attr_filters: dict: Optional dictionary of Attribute filters
        :param api_version: Optional API version to use for this request's X-API-Version header,
                            default None will use latest version.
        :return: int: Count
        """
        return self.client.get_count(
            self.endpoint, filters=filters, attr_filters=attr_filters, snapshot=False, api_version=api_version
        )


class Table(BaseTable, BaseModel):
    """model for table data"""

    sn: Optional[str] = Field(None, exclude=None)

    def _dev_filters(self, url, columns=None, filters=None):
        if self.sn:
            all_columns, f = create_filter(self.client, self.client._check_url(url), filters=filters, sn=self.sn)
            return columns or all_columns, f
        return columns, filters

    @overload
    def fetch(
        self,
        export: Literal["json"] = ...,
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        api_version: Optional[Union[str, int]] = None,
    ) -> list[dict]: ...

    @overload
    def fetch(
        self,
        export: Literal["csv"],
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        csv_tz: Optional[str] = None,
        # api_version: Optional[Union[str, int]] = None,  # TODO: NIM-21720
    ) -> bytes: ...

    @overload
    def fetch(
        self,
        export: Literal["df"],
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        api_version: Optional[Union[str, int]] = None,
    ) -> DataFrame: ...

    def fetch(
        self,
        export: EXPORT_FORMAT = "json",
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        csv_tz: Optional[str] = None,
        api_version: Optional[Union[str, int]] = None,
    ):
        """Gets all data from corresponding endpoint

        Args:
            export: str: Export format to return [json, csv]; default is json.
            columns: Optional columns to return, default is all
            filters: Optional filters'
            attr_filters: dict: Optional dictionary of Attribute filters
            snapshot_id: Optional snapshot ID to override class
            reports: True to return Intent Rules (also accepts string of frontend URL) or a list of report IDs
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            limit: Default to 1,000 rows
            start: Starts at 0
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
            api_version: Optional API version to use for this request's X-API-Version header,
                         default None will use latest version. Values other than None will not use streaming requests
                         and will switch to pagination. API Version is not supported with CSV export.
        Returns:
            Union[list[dict], bytes, pd.DataFrame]: List of dict if json, bytes string if CSV, DataFrame if df
        """
        columns, filters = self._dev_filters(self.endpoint, columns, filters)
        return self.client.fetch(
            self.endpoint,
            export=export,
            columns=columns,
            filters=filters,
            attr_filters=attr_filters,
            snapshot_id=snapshot_id,
            reports=reports,
            sort=sort,
            limit=limit,
            start=start,
            csv_tz=csv_tz,
            api_version=api_version,
        )

    @overload
    def all(
        self,
        export: Literal["json"] = ...,
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        api_version: Optional[Union[str, int]] = None,
    ) -> list[dict]: ...

    @overload
    def all(
        self,
        export: Literal["csv"],
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        csv_tz: Optional[str] = None,
        # api_version: Optional[Union[str, int]] = None,  # TODO: NIM-21720
    ) -> bytes: ...

    @overload
    def all(
        self,
        export: Literal["df"],
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        api_version: Optional[Union[str, int]] = None,
    ) -> DataFrame: ...

    def all(
        self,
        export: EXPORT_FORMAT = "json",
        columns: Optional[Union[list[str], set[str]]] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        csv_tz: Optional[str] = None,
        api_version: Optional[Union[str, int]] = None,
    ):
        """Gets all data from corresponding endpoint

        Args:
            export: str: Export format to return [json, csv]; default is json.
            columns: Optional columns to return, default is all
            filters: Optional filters
            attr_filters: dict: Optional dictionary of Attribute filters
            snapshot_id: Optional snapshot ID to override class
            reports: True to return Intent Rules (also accepts string of frontend URL) or a list of report IDs
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
            api_version: Optional API version to use for this request's X-API-Version header,
                         default None will use latest version. Values other than None will not use streaming requests
                         and will switch to pagination. API Version is not supported with CSV export.
        Returns:
            Union[list[dict], bytes, pd.DataFrame]: List of dict if json, bytes string if CSV, DataFrame if df
        """
        columns, filters = self._dev_filters(self.endpoint, columns, filters)
        return self.client.fetch_all(
            self.endpoint,
            export=export,
            columns=columns,
            filters=filters,
            attr_filters=attr_filters,
            snapshot_id=snapshot_id,
            reports=reports,
            sort=sort,
            csv_tz=csv_tz,
            api_version=api_version,
        )

    def count(
        self,
        filters: Optional[dict] = None,
        snapshot_id: Optional[str] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        api_version: Optional[Union[str, int]] = None,
    ) -> int:
        """
        Gets count of table
        Args:
            filters: dict: Optional filters
            snapshot_id: str: Optional snapshot ID to override class
            attr_filters: dict: Optional dictionary of Attribute filters
            api_version: Optional API version to use for this request's X-API-Version header, default latest.

        Returns:
            int: Count of rows
        """
        _, filters = self._dev_filters(self.endpoint, filters=filters)
        return self.client.get_count(
            self.endpoint, filters=filters, attr_filters=attr_filters, snapshot_id=snapshot_id, api_version=api_version
        )

    @staticmethod
    def _ignore_columns(columns: set, columns_ignore: set):
        """
        Determines which columns to use in the query.
        Args:
            columns: set : Set of columns to use
            columns_ignore: set : Set of columns to ignore

        Returns:
            set[str]: Set of columns to use
        """
        return {col for col in columns if col not in columns_ignore or col == "id"}

    def _compare_determine_columns(self, columns: set, columns_ignore: set, unique_keys: set, nested_keys: set):
        """
        Determines which columns to use in the query, and which columns to use
        Only supports a single nested column
        Args:
            columns: set : Set of columns to use
            columns_ignore: set : Set of columns to ignore
            unique_keys: set : Set of columns for unique keys
            nested_keys: set: Set of nested columns that belongs to a column in columns, columns_ignore, or unique_keys

        Returns:
            tuple[list, list]: List of columns to use in query, List of columns to use when sorting data
        """
        # get all columns for the table
        table_columns = set(self.client.get_columns(self.endpoint))
        nested_cols_for_return = (
            {col for col in set(self.client.oas[self.endpoint].post.nested_columns) if col in nested_keys}
            if nested_keys
            else None
        )

        # Must always ignore some columns
        columns_ignore.update(IGNORE_COLUMNS)

        # user passes unique_keys
        if unique_keys:
            if not table_columns.issuperset(unique_keys):
                raise ValueError(f"Unique Key(s) {unique_keys - table_columns} not in table {self.name}")
            return list(unique_keys), nested_cols_for_return
        # user passes columns
        if columns:
            if not table_columns.issuperset(columns):
                raise ValueError(f"Column(s) {columns - table_columns} not in table {self.name}")
            return list(self._ignore_columns(columns, columns_ignore)), nested_cols_for_return
        # user does not pass columns
        # or only passed columns_ignore
        return list(self._ignore_columns(table_columns, columns_ignore)), nested_cols_for_return

    @staticmethod
    def _remove_keys_from_dict(dict_object, keys_to_remove):
        if not keys_to_remove:
            return dict_object

        for v in dict_object.values():
            if not isinstance(v, list):
                continue

            for item in v:
                for key in keys_to_remove:
                    if key in item:
                        item.pop(key)

        return dict_object

    def _hash_data(self, json_data, unique_keys=None, nested_columns_ignore=None):
        """
        Hashes data. Turns any data into a string and hashes it, then returns the hash as a key for the data
        Args:
            json_data: list[dict] : List of dictionaries to hash
            unique_keys: list[str] : List of keys to use for hashing
            nested_columns_ignore: list[str]: List of nested column keys to filter data

        Returns:
            dict[str]: dictionary with hash as key and values as the original data
        """
        # loop over each obj, turn the obj into a string, and hash it
        return_json = {}
        # user passes unique_keys
        if unique_keys:
            # loop over each response
            for dict_obj in json_data:
                dict_for_hash = self._remove_keys_from_dict(dict_object=dict_obj, keys_to_remove=nested_columns_ignore)
                # create a dictionary, of only the keys/columns requested
                hash_key = {key: dict_for_hash[key] for key in unique_keys}
                # hash the data
                unique_hash = deepdiff.DeepHash(hash_key)[hash_key]
                # check if the data has already been processed.
                if unique_hash in return_json:
                    # raise error if data has already been processed
                    raise KeyError(f"Unique Key(s) {unique_keys} are not unique, please adjust unique_keys input.")
                # store the data, using the hash as a key
                return_json[unique_hash] = dict_for_hash
        # user passes columns and/or column ignore
        else:
            # loop over each response
            for dict_obj in json_data:
                dict_for_hash = self._remove_keys_from_dict(dict_object=dict_obj, keys_to_remove=nested_columns_ignore)
                # hash each object, and store it in a dictionary, with the hash as the key
                return_json[deepdiff.DeepHash(dict_for_hash)[dict_for_hash]] = dict_for_hash
        return return_json

    @staticmethod
    def _make_set(data: Union[list, set, str] = None):
        if isinstance(data, str):
            return {data}
        elif data is None:
            return set()
        else:
            return set(data)

    def compare(
        self,
        snapshot_id: str = None,
        columns: Union[list, set] = None,
        columns_ignore: Union[list, set, str] = None,
        unique_keys: Union[list, set, str] = None,
        nested_columns_ignore: Union[list, set, str] = None,
        **kwargs,
    ):
        """
        Compares a table from the current snapshot to the snapshot_id passed.
        Args:
            snapshot_id: str : The snapshot_id to compare to.
            columns: list : List of columns to compare. If None, will compare all columns.
            columns_ignore: list : List of columns to ignore. If None, will always ignore 'id' column.
            unique_keys: list : List of columns to use as unique keys. If None, will use all columns as primary key.
            nested_columns_ignore: List of columns that belong to a nested columns
            **kwargs: dict : Optional Table.all() arguments to apply to the table before comparing.

        Returns:
            dict : dictionary containing the differences between the two snapshots.
                   Possible keys are 'added', 'removed' and 'changed'.
        """
        return_dict = {}

        # determine which columns to use in query
        columns = self._make_set(columns)
        columns_ignore = self._make_set(columns_ignore)
        unique_keys = self._make_set(unique_keys)
        nested_columns_ignore = self._make_set(nested_columns_ignore)
        cols_for_query, nested_cols = self._compare_determine_columns(
            columns, columns_ignore, unique_keys, nested_columns_ignore
        )
        data = self.all(columns=cols_for_query, **kwargs)
        data_compare = self.all(snapshot_id=snapshot_id, columns=cols_for_query, **kwargs)

        # since we turned the values into a hash, we can just compare the keys
        if unique_keys:
            hashed_data_unique = self._hash_data(data, unique_keys, nested_cols)
            hashed_data_compare_unique = self._hash_data(data_compare, unique_keys, nested_cols)
            changed = [
                hashed_data_unique[hashed_str]
                for hashed_str in hashed_data_unique.keys()
                if hashed_str not in hashed_data_compare_unique.keys()
            ]
            return_dict["changed"] = changed
            return return_dict
        # compare both ways
        hashed_data = self._hash_data(data, nested_columns_ignore=nested_cols)
        hashed_data_compare = self._hash_data(data_compare, nested_columns_ignore=nested_cols)
        added = [
            hashed_data[hashed_str] for hashed_str in hashed_data.keys() if hashed_str not in hashed_data_compare.keys()
        ]
        removed = [
            hashed_data_compare[hashed_str]
            for hashed_str in hashed_data_compare.keys()
            if hashed_str not in hashed_data.keys()
        ]
        return_dict["added"] = added
        return_dict["removed"] = removed
        return return_dict

    def join_table(
        self,
        on_column: Union[str, dict[str, str], list[str], tuple[str], set[str]],
        join_table_path: str,
        how: Literal["right", "left"] = "right",
    ) -> Dict:
        """
        Performs a join operation on on_columns, assuming on_columns are columns that
        exist in both self and join_table_path. If on_columns are not columns in
        set(self.client.get_columns(self.endpoint)), ValueError is raised.

        Args:
            on_column: Union[str, dict[str, str], list[str], tuple[str], set[str]] : column(s) to use for join operation
            join_table_path: URL of table to use for join operation
            how: str : type of join operation ('right' or 'left')

        Returns: dict: Dictionary of result of joined table.
            possible keys <join_table_path>, self.endpoint, joined, no_matches
        """
        if isinstance(on_column, str):
            on_column = {on_column: on_column}
        elif isinstance(on_column, (list, set, tuple)):
            on_column = {col: col for col in on_column}

        self_cols = set(self.client.get_columns(self.endpoint))
        join_table_path = self.client._check_url(join_table_path)
        join_cols = set(self.client.get_columns(join_table_path))

        for col_self, col_join in on_column.items():
            if col_self not in self_cols:
                raise ValueError(f"Column {col_self} is not in self table columns {self_cols}.")
            if col_join not in join_cols:
                raise ValueError(f"Column {col_join} is not in join table columns {join_cols}.")

        self_table = self.client.fetch_all(self.endpoint)
        join_table = self.client.fetch_all(join_table_path)

        def match_rows(self_dict, join_dict):
            return all(self_dict[col_self] == join_dict[col_join] for col_self, col_join in on_column.items())

        if how == "right":
            joined_data = [
                {**self_dict, **join_dict}
                for join_dict in join_table
                for self_dict in self_table
                if match_rows(self_dict, join_dict)
            ]
            no_matches = [
                self_dict
                for self_dict in self_table
                if not any(match_rows(self_dict, join_dict) for join_dict in join_table)
            ]
        else:
            joined_data = [
                {**join_dict, **self_dict}
                for self_dict in self_table
                for join_dict in join_table
                if match_rows(self_dict, join_dict)
            ]
            no_matches = [
                join_dict
                for join_dict in join_table
                if not any(match_rows(self_dict, join_dict) for self_dict in self_table)
            ]

        return {
            "joined": joined_data,
            self.endpoint: self_table,
            join_table_path: join_table,
            "no_matches": no_matches,
        }



from datetime import datetime
from typing import Optional, Literal, Annotated

from pydantic import BaseModel, Field, AliasPath, BeforeValidator, AliasChoices

WEBHOOK_TYPES = Literal["snapshot", "intent-verification"]


class BaseEvent(BaseModel):
    test: Optional[bool] = Field(False, description="Indicates if this is a test event.")
    reason: Optional[str] = None
    requester: str = Field(examples=["cron", "user:<id>", "snapshot:<action>", "recalculateSites"])
    timestamp: datetime = Field(description="Timestamp of the event.")
    status: str = Field(examples=["started", "completed", "failed", "resumed", "resumed (stopping)", "stopped"])


class SnapshotEvent(BaseEvent, BaseModel):
    type: Literal["snapshot"] = "snapshot"
    action: Literal["discover", "clone", "delete", "download", "load", "unload"] = Field(
        examples=["discover", "clone", "delete", "download", "load", "unload"]
    )
    snapshotId: Optional[str] = Field(None, validation_alias=AliasChoices("snapshotId", AliasPath("snapshot", "id")))
    name: Optional[str] = Field(
        None,
        validation_alias=AliasChoices("name", AliasPath("snapshot", "name")),
        description="Name of the snapshot.",
    )
    cloneId: Optional[str] = Field(
        None,
        validation_alias=AliasChoices("cloneId", AliasPath("snapshot", "cloneId")),
        description="ID of the snapshot being cloned.",
    )
    file: Optional[str] = Field(
        None,
        validation_alias=AliasChoices(AliasPath("snapshot", "file")),
        description="Filename associated with the snapshot download.",
    )


class IntentEvent(BaseEvent, BaseModel):
    type: Literal["intent-verification"] = "intent-verification"
    action: Literal["calculate"] = "calculate"
    reportId: Optional[Annotated[str, BeforeValidator(str)]] = None
    snapshotId: Optional[str] = None


try:
    from sqlmodel import Field as SField, SQLModel

    class WebhookEvent(SnapshotEvent, IntentEvent, SQLModel, table=True):
        __tablename__ = "webhook_events"
        dbId: Optional[int] = SField(default=None, primary_key=True, description="Auto-generated database ID")
        snapshotId: Optional[str] = SField(None, index=True)
        type: str = SField(index=True, description="One of 'intent-verification' or 'snapshot'")
        action: str = Field(examples=["discover", "clone", "delete", "download", "load", "unload", "calculate"])

except ImportError:
    WebhookEvent = None



from datetime import datetime
from typing import Optional, Any, Union
from uuid import uuid4

from pydantic import BaseModel, Field, field_validator, model_validator, model_serializer, RootModel

from ipfabric.tools.shared import VALID_IP, VALID_IPv6, validate_ip_network_str


class Expiration(BaseModel):
    enabled: bool = False
    value: Optional[datetime] = None

    @model_validator(mode="after")
    def _verify(self):
        if self.enabled and not self.value:
            raise ValueError("Enabled Expiration requires a value.")
        elif not self.enabled and self.value:
            raise ValueError("Expiration value requires enabled to be True.")
        return self

    @model_serializer
    def _ser_model(self) -> dict[str, Any]:
        return (
            {"enabled": self.enabled, "value": self.value.strftime("%Y-%m-%d %H:%M:%S")}
            if self.enabled
            else {"enabled": self.enabled}
        )


class BaseCred(BaseModel):
    custom: bool = False
    username: str
    network: list[Union[str, VALID_IP, VALID_IPv6]]
    excluded: Optional[list[Union[str, VALID_IP, VALID_IPv6]]] = Field(default_factory=list, alias="excludeNetworks")
    expiration: Expiration = Field(None, alias="expirationDate")
    encrypt_password: str = Field(None, alias="password")
    priority: Optional[int] = None
    notes: Optional[str] = None

    @field_validator("network", "excluded")
    @classmethod
    def _verify_valid_networks(cls, v: list[str]) -> list[str]:
        return [validate_ip_network_str(_, ipv6=True) for _ in v]

    @model_serializer
    def _ser_model(self) -> dict[str, Any]:
        custom = {"custom": self.custom} if self.custom else {}
        return dict(
            priority=self.priority,
            username=self.username,
            excludeNetworks=self.excluded,
            expirationDate=self.expiration.model_dump(),
            password=self.encrypt_password,
            notes=self.notes or "",
            **custom,
        )


class Credential(BaseCred, BaseModel):
    credential_id: Optional[str] = Field(None, alias="id")
    config_mgmt: bool = Field(False, alias="syslog")

    @model_validator(mode="after")
    def _verify(self):
        if not self.credential_id:
            self.custom = True
        return self

    @model_serializer
    def _ser_cred_model(self) -> dict[str, Any]:
        cred = self._ser_model()
        return dict(
            syslog=self.config_mgmt,
            id=self.credential_id if self.credential_id else str(uuid4()),
            network=self.network,
            **cred,
        )


class Privilege(BaseCred, BaseModel):
    privilege_id: str = Field(None, alias="id")
    network: list[Union[str, VALID_IP]] = Field(None, alias="includeNetworks")

    @model_validator(mode="after")
    def _verify(self):
        if not self.privilege_id:
            self.custom = True
        return self

    @model_serializer
    def _ser_cred_model(self) -> dict[str, Any]:
        cred = self._ser_model()
        return dict(id=self.privilege_id if self.privilege_id else str(uuid4()), includeNetworks=self.network, **cred)


def serialize_list(auth):
    if not auth:
        return []
    priorities = sorted([_.priority for _ in auth])
    if priorities != list(range(1, len(priorities) + 1)):
        raise ValueError("Priorities are not sequential.")
    return [_.model_dump() for _ in auth]


class CredentialList(RootModel):
    root: list[Credential]

    @model_serializer
    def _verify_priorities(self) -> dict[str, Any]:
        return serialize_list(self.credentials)

    def __bool__(self):
        return bool(self.credentials)

    def __getattr__(self, name: str) -> Any:
        return getattr(self.root, name)


class PrivilegeList(RootModel):
    root: list[Privilege]

    @model_serializer
    def _verify_priorities(self) -> dict[str, Any]:
        return serialize_list(self.privileges)

    def __getattr__(self, name: str) -> Any:
        return getattr(self.root, name)
